@@ -36,12 +36,13 @@
 typedef struct {
     zsock_t *pipe;              //  Actor command pipe
     SOCKET udpsock;             //  UDP socket for send/recv
-    int port_nbr;               //  UDP port number we work on
+    SOCKET udpsock_send;        //  UDP socket for IPv6 send
+    char port_nbr [7];          //  UDP port number we work on
     int interval;               //  Beacon broadcast interval
     int64_t ping_at;            //  Next broadcast time
     zframe_t *transmit;         //  Beacon transmit data
     zframe_t *filter;           //  Beacon filter data
-    inaddr_t broadcast;         //  Our broadcast address
+    inaddr_storage_t broadcast; //  Our broadcast address
     bool terminated;            //  Did caller ask us to quit?
     bool verbose;               //  Verbose logging enabled?
     char hostname [NI_MAXHOST]; //  Saved host name
@@ -57,7 +58,7 @@ s_self_destroy (self_t **self_p)
         zframe_destroy (&self->filter);
         if (self->udpsock) // don't close STDIN
             zsys_udp_close (self->udpsock);
-        free (self);
+        freen (self);
         *self_p = NULL;
     }
 }
@@ -81,37 +82,150 @@ s_self_prepare_udp (self_t *self)
     //  Create our UDP socket
     if (self->udpsock)
         zsys_udp_close (self->udpsock);
+    if (self->udpsock_send)
+        zsys_udp_close (self->udpsock_send);
 
     self->hostname [0] = 0;
+    //  For IPv6 we need two sockets. At least on Linux, IPv6 multicast packets
+    //  are NOT received despite joining the group and setting the interface
+    //  option UNLESS the socket is bound to in6_addrany, which means the kernel
+    //  will select an arbitrary IP address as the source when sending beacons
+    //  out. This breaks zbeacon as the protocol uses the source address of a
+    //  beacon to find the endpoint of a peer, which is then random and
+    //  useless (could even be associated with a different interface, eg: a
+    //  virtual bridge).
+    //  As a workaround, use a different socket to send packets. So the socket
+    //  that receives can be bound to in6_addrany, and the socket that sends
+    //  can be bound to the actual intended host address.
     self->udpsock = zsys_udp_new (false);
-    if (self->udpsock == INVALID_SOCKET)
+    if (self->udpsock == INVALID_SOCKET) {
+        self->udpsock_send = INVALID_SOCKET;
         return;
+    }
+    self->udpsock_send = zsys_udp_new (false);
+    if (self->udpsock_send == INVALID_SOCKET) {
+        zsys_udp_close (self->udpsock);
+        self->udpsock = INVALID_SOCKET;
+        return;
+    }
 
     //  Get the network interface fro ZSYS_INTERFACE or else use first
     //  broadcast interface defined on system. ZSYS_INTERFACE=* means
     //  use INADDR_ANY + INADDR_BROADCAST.
     const char *iface = zsys_interface ();
-    in_addr_t bind_to = 0;
-    in_addr_t send_to = 0;
+    struct addrinfo *bind_to = NULL;
+    struct addrinfo *send_to = NULL;
+    struct addrinfo hint;
+    memset (&hint, 0, sizeof(struct addrinfo));
+    hint.ai_flags = AI_NUMERICHOST;
+#if !defined (CZMQ_HAVE_ANDROID) && !defined (CZMQ_HAVE_FREEBSD)
+    hint.ai_flags |= AI_V4MAPPED;
+#endif
+    hint.ai_socktype = SOCK_DGRAM;
+    hint.ai_protocol = IPPROTO_UDP;
+    hint.ai_family = zsys_ipv6 () ? AF_INET6 : AF_INET;
+    int rc;
     int found_iface = 0;
+    unsigned int if_index = 0;
 
     if (streq (iface, "*")) {
         //  Wildcard means bind to INADDR_ANY and send to INADDR_BROADCAST
-        bind_to = INADDR_ANY;
-        send_to = INADDR_BROADCAST;
+        //  IE - getaddrinfo with NULL as first parameter or 255.255.255.255
+        //  (or IPv6 multicast link-local all-node group ff02::1
+        hint.ai_flags = hint.ai_flags | AI_PASSIVE;
+        rc = getaddrinfo (NULL, self->port_nbr, &hint, &bind_to);
+        assert (rc == 0);
+
+        if (zsys_ipv6()) {
+            //  Default is link-local all-node multicast group
+            rc = getaddrinfo (zsys_ipv6_mcast_address (), self->port_nbr, &hint,
+                    &send_to);
+            assert (rc == 0);
+        }
+        else {
+            rc = getaddrinfo ("255.255.255.255", self->port_nbr, &hint,
+                    &send_to);
+            assert (rc == 0);
+        }
+
+        found_iface = 1;
+    }
+    // if ZSYS_INTERFACE is a single digit, use the corresponding interface in
+    // the interface list
+    else if (strlen (iface) == 1 && iface[0] >= '0' && iface[0] <= '9')
+    {
+        int if_number = atoi (iface);
+        ziflist_t *iflist = ziflist_new_ipv6 ();
+        assert (iflist);
+        const char *name = ziflist_first (iflist);
+        int idx = -1;
+        while (name) {
+            idx++;
+            if (idx == if_number &&
+                    ((ziflist_is_ipv6 (iflist) && zsys_ipv6 ()) ||
+                            (!ziflist_is_ipv6 (iflist) && !zsys_ipv6 ()))) {
+                //  Using inet_addr instead of inet_aton or inet_atop
+                //  because these are not supported in Win XP
+                rc = getaddrinfo (ziflist_address (iflist), self->port_nbr,
+                        &hint, &bind_to);
+                assert (rc == 0);
+                rc = getaddrinfo (ziflist_broadcast (iflist), self->port_nbr,
+                        &hint, &send_to);
+                assert (rc == 0);
+                if_index = if_nametoindex (name);
+
+                if (self->verbose)
+                    zsys_info ("zbeacon: interface=%s address=%s broadcast=%s",
+                            name, ziflist_address (iflist), ziflist_broadcast (iflist));
+                found_iface = 1;
+                break;      //  iface is known, so allow it
+            }
+            name = ziflist_next (iflist);
+        }
+        ziflist_destroy (&iflist);
+    }
+    else if (zsys_ipv6 () && strneq("", zsys_ipv6_address ()) && strneq (iface, "")) {
+        rc = getaddrinfo (zsys_ipv6_address (), self->port_nbr,
+                &hint, &bind_to);
+        assert (rc == 0);
+        //  A user might set a link-local address without appending %iface
+        if (IN6_IS_ADDR_LINKLOCAL (&((in6addr_t *)bind_to->ai_addr)->sin6_addr) &&
+                !strchr (zsys_ipv6_address (), '%')) {
+            char address_and_iface [NI_MAXHOST] = {0};
+            strcat (address_and_iface, zsys_ipv6_address ());
+            strcat (address_and_iface, "%");
+            strcat (address_and_iface, iface);
+            rc = getaddrinfo (address_and_iface, self->port_nbr, &hint, &bind_to);
+            assert (rc == 0);
+        }
+        rc = getaddrinfo (zsys_ipv6_mcast_address (), self->port_nbr,
+                &hint, &send_to);
+        assert (rc == 0);
+        if_index = if_nametoindex (iface);
+
+        if (self->verbose)
+            zsys_info ("zbeacon: interface=%s address=%s broadcast=%s",
+                    iface, zsys_ipv6_address (), zsys_ipv6_mcast_address ());
         found_iface = 1;
     }
     else {
         //  Look for matching interface, or first ziflist item
-        ziflist_t *iflist = ziflist_new ();
+        ziflist_t *iflist = ziflist_new_ipv6 ();
         assert (iflist);
         const char *name = ziflist_first (iflist);
         while (name) {
-            if (streq (iface, name) || streq (iface, "")) {
-                //  Using inet_addr instead of inet_aton or inet_atop
-                //  because these are not supported in Win XP
-                send_to = inet_addr (ziflist_broadcast (iflist));
-                bind_to = inet_addr (ziflist_address (iflist));
+            //  If IPv6 is not enabled ignore IPv6 interfaces.
+            if ((streq (iface, name) || streq (iface, "")) &&
+                    ((ziflist_is_ipv6 (iflist) && zsys_ipv6 ()) ||
+                            (!ziflist_is_ipv6 (iflist) && !zsys_ipv6 ()))) {
+                rc = getaddrinfo (ziflist_address (iflist), self->port_nbr,
+                        &hint, &bind_to);
+                assert (rc == 0);
+                rc = getaddrinfo (ziflist_broadcast (iflist), self->port_nbr,
+                        &hint, &send_to);
+                assert (rc == 0);
+                if_index = if_nametoindex (name);
+
                 if (self->verbose)
                     zsys_info ("zbeacon: interface=%s address=%s broadcast=%s",
                                name, ziflist_address (iflist), ziflist_broadcast (iflist));
@@ -123,37 +237,63 @@ s_self_prepare_udp (self_t *self)
         ziflist_destroy (&iflist);
     }
     if (found_iface) {
-        self->broadcast.sin_family = AF_INET;
-        self->broadcast.sin_port = htons (self->port_nbr);
-        self->broadcast.sin_addr.s_addr = send_to;
-        inaddr_t address = self->broadcast;
-        address.sin_addr.s_addr = bind_to;
-        //  Bind to the port on all interfaces
+        inaddr_storage_t bind_address;
+
+        //  On Windows we bind to the host address
+        //  On *NIX we bind to INADDR_ANY or in6addr_any, otherwise multicast
+        //  packets will be filtered out despite joining the group
 #if (defined (__WINDOWS__))
-        inaddr_t sockaddr = address;
-#elif (defined (__APPLE__))
-        inaddr_t sockaddr = self->broadcast;
-        sockaddr.sin_addr.s_addr = htons (INADDR_ANY);
+        memcpy (&bind_address, bind_to->ai_addr, bind_to->ai_addrlen);
 #else
-        inaddr_t sockaddr = self->broadcast;
+        memcpy (&bind_address, send_to->ai_addr, send_to->ai_addrlen);
+        if (zsys_ipv6 ())
+            bind_address.__inaddr_u.__addr6.sin6_addr = in6addr_any;
+        else
+            bind_address.__inaddr_u.__addr.sin_addr.s_addr = htonl (INADDR_ANY);
 #endif
+        memcpy (&self->broadcast, send_to->ai_addr, send_to->ai_addrlen);
+
+        if (zsys_ipv6()) {
+            struct ipv6_mreq mreq;
+            mreq.ipv6mr_interface = if_index;
+            memcpy (&mreq.ipv6mr_multiaddr,
+                    &(((in6addr_t *)(send_to->ai_addr))->sin6_addr),
+                    sizeof (struct in6_addr));
+
+            if (setsockopt (self->udpsock, IPPROTO_IPV6, IPV6_JOIN_GROUP,
+                    (char *)&mreq, sizeof (mreq)))
+                zsys_socket_error ("zbeacon: setsockopt IPV6_JOIN_GROUP failed");
+
+            if (setsockopt (self->udpsock, IPPROTO_IPV6, IPV6_MULTICAST_IF,
+                    (char *)&if_index, sizeof (if_index)))
+                zsys_socket_error ("zbeacon: setsockopt IPV6_MULTICAST_IF failed");
+
+            if (setsockopt (self->udpsock_send, IPPROTO_IPV6, IPV6_JOIN_GROUP,
+                    (char *)&mreq, sizeof (mreq)))
+                zsys_socket_error ("zbeacon: setsockopt IPV6_JOIN_GROUP failed");
+
+            if (setsockopt (self->udpsock_send, IPPROTO_IPV6, IPV6_MULTICAST_IF,
+                    (char *)&if_index, sizeof (if_index)))
+                zsys_socket_error ("zbeacon: setsockopt IPV6_MULTICAST_IF failed");
+        }
+
         //  If bind fails, we close the socket for opening again later (next poll interval)
-        if (bind (self->udpsock, (struct sockaddr *) &sockaddr, sizeof (inaddr_t)))
-        {
+        if (bind (self->udpsock_send, bind_to->ai_addr, bind_to->ai_addrlen) ||
+                bind (self->udpsock, (struct sockaddr *)&bind_address,
+                zsys_ipv6 () ? sizeof (in6addr_t) : sizeof (inaddr_t))) {
             zsys_debug ("zbeacon: Unable to bind to broadcast address, reason=%s", strerror (errno));
             zsys_udp_close (self->udpsock);
             self->udpsock = INVALID_SOCKET;
-            return;
+            zsys_udp_close (self->udpsock_send);
+            self->udpsock_send = INVALID_SOCKET;
         }
-
-        //  Get our hostname so we can send it back to the API
-        if (address.sin_addr.s_addr == INADDR_ANY) {
+        else if (streq (iface, "*")) {
             strcpy(self->hostname, "*");
             if (self->verbose)
                 zsys_info ("zbeacon: configured, hostname=%s", self->hostname);
         }
-        else if (getnameinfo ((struct sockaddr *) &address, sizeof (inaddr_t),
-                              self->hostname, NI_MAXHOST, NULL, 0, NI_NUMERICHOST) == 0) {
+        else if (getnameinfo (bind_to->ai_addr, bind_to->ai_addrlen,
+                        self->hostname, NI_MAXHOST, NULL, 0, NI_NUMERICHOST) == 0) {
             if (self->verbose)
                 zsys_info ("zbeacon: configured, hostname=%s", self->hostname);
         }
@@ -163,7 +303,12 @@ s_self_prepare_udp (self_t *self)
         //  No valid interface. Close the socket so that we can try again later
         zsys_udp_close(self->udpsock);
         self->udpsock = INVALID_SOCKET;
+        zsys_udp_close (self->udpsock_send);
+        self->udpsock_send = INVALID_SOCKET;
     }
+
+    freeaddrinfo (bind_to);
+    freeaddrinfo (send_to);
 }
 
 
@@ -175,7 +320,7 @@ static void
 s_self_configure (self_t *self, int port_nbr)
 {
     assert (port_nbr);
-    self->port_nbr = port_nbr;
+    snprintf (self->port_nbr, 7, "%d", port_nbr);
     s_self_prepare_udp (self);
     zstr_send (self->pipe, self->hostname);
     if (streq (self->hostname, ""))
@@ -248,8 +393,10 @@ s_self_handle_udp (self_t *self)
 {
     assert (self);
 
-    char peername [INET_ADDRSTRLEN];
-    zframe_t *frame = zsys_udp_recv (self->udpsock, peername, INET_ADDRSTRLEN);
+    char peername [NI_MAXHOST];
+    zframe_t *frame = zsys_udp_recv (self->udpsock, peername, NI_MAXHOST);
+    if (!frame)
+        return;
 
     //  If filter is set, check that beacon matches it
     bool is_valid = false;
@@ -282,6 +429,52 @@ s_self_handle_udp (self_t *self)
 }
 
 
+//  --------------------------------------------------------------------------
+//  Send the beacon over UDP
+
+static int
+s_emit_beacon (self_t *self)
+{
+#if defined (__WINDOWS__)
+    // Windows doesn't broadcast on all interfaces when using INADDR_BROADCAST
+    // only the interface with the highest metric (as seen in `route print`)
+    // so send a packet per interface to each broadcast address
+    if (streq (zsys_interface (), "*") && !zsys_ipv6 ()) {
+        INTERFACE_INFO interface_list [64];
+        DWORD bytes_received = 0;
+
+        int rc = WSAIoctl (self->udpsock, SIO_GET_INTERFACE_LIST, 0, 0,
+            &interface_list, sizeof (interface_list), &bytes_received, NULL, NULL);
+        assert (rc != SOCKET_ERROR);
+
+        int num_interfaces = bytes_received / sizeof (INTERFACE_INFO);
+
+        // iiBroadcastAddress is always 255.255.255.255 need to calculate the specific broadcast address using the netmask
+        // keep the same parameters as self->broadcast but just replace the address for each interface
+        inaddr_t addr;
+        memcpy(&addr, &self->broadcast, sizeof (inaddr_t));
+
+        for (int i = 0; i < num_interfaces; ++i)
+        {
+            addr.sin_addr.S_un.S_addr = (interface_list[i].iiAddress.AddressIn.sin_addr.S_un.S_addr
+                | ~(interface_list[i].iiNetmask.AddressIn.sin_addr.S_un.S_addr));
+
+            if (zsys_udp_send (self->udpsock_send, self->transmit,
+                (inaddr_t *)&addr, sizeof (inaddr_t))) {
+                // Send failed, cause zbeacon to re-init socket
+                return -1;
+            }
+        }
+        return 0;
+    }
+#endif
+
+    return zsys_udp_send (self->udpsock_send, self->transmit,
+        (inaddr_t *)&self->broadcast,
+        zsys_ipv6 () ? sizeof (in6addr_t) : sizeof (inaddr_t));
+}
+
+
 //  --------------------------------------------------------------------------
 //  zbeacon() implements the zbeacon actor interface
 
@@ -317,10 +510,9 @@ zbeacon (zsock_t *pipe, void *args)
         if (self->transmit
         &&  zclock_mono () >= self->ping_at) {
             //  Send beacon to any listening peers
-            if (!self->udpsock || self->udpsock == INVALID_SOCKET ||
-                zsys_udp_send (self->udpsock, self->transmit, &self->broadcast, sizeof(inaddr_t)))
+            if (!self->udpsock_send || self->udpsock_send == INVALID_SOCKET || s_emit_beacon(self))
             {
-                const char *reason = (!self->udpsock || self->udpsock == INVALID_SOCKET) ? "invalid socket" : strerror (errno);
+                const char *reason = (!self->udpsock_send || self->udpsock_send == INVALID_SOCKET) ? "invalid socket" : strerror (errno);
                 zsys_debug ("zbeacon: failed to transmit, attempting reconnection. reason=%s", reason);
                 //  Try to recreate UDP socket on interface
                 s_self_prepare_udp (self);
@@ -355,10 +547,10 @@ zbeacon_test (bool verbose)
     if (!*hostname) {
         printf ("OK (skipping test, no UDP broadcasting)\n");
         zactor_destroy (&speaker);
-        free (hostname);
+        freen (hostname);
         return;
     }
-    free (hostname);
+    freen (hostname);
 
     //  Create listener beacon on port 9999 to lookup service
     zactor_t *listener = zactor_new (zbeacon, NULL);
@@ -368,7 +560,7 @@ zbeacon_test (bool verbose)
     zsock_send (listener, "si", "CONFIGURE", 9999);
     hostname = zstr_recv (listener);
     assert (*hostname);
-    free (hostname);
+    freen (hostname);
 
     //  We will broadcast the magic value 0xCAFE
     byte announcement [2] = { 0xCA, 0xFE };
@@ -397,21 +589,21 @@ zbeacon_test (bool verbose)
     zsock_send (node1, "si", "CONFIGURE", 5670);
     hostname = zstr_recv (node1);
     assert (*hostname);
-    free (hostname);
+    freen (hostname);
 
     zactor_t *node2 = zactor_new (zbeacon, NULL);
     assert (node2);
     zsock_send (node2, "si", "CONFIGURE", 5670);
     hostname = zstr_recv (node2);
     assert (*hostname);
-    free (hostname);
+    freen (hostname);
 
     zactor_t *node3 = zactor_new (zbeacon, NULL);
     assert (node3);
     zsock_send (node3, "si", "CONFIGURE", 5670);
     hostname = zstr_recv (node3);
     assert (*hostname);
-    free (hostname);
+    freen (hostname);
 
     zsock_send (node1, "sbi", "PUBLISH", "NODE/1", 6, 250);
     zsock_send (node2, "sbi", "PUBLISH", "NODE/2", 6, 250);
@@ -450,6 +642,10 @@ zbeacon_test (bool verbose)
     zactor_destroy (&node1);
     zactor_destroy (&node2);
     zactor_destroy (&node3);
+
+#if defined (__WINDOWS__)
+    zsys_shutdown();
+#endif
     //  @end
     printf ("OK\n");
 }
