@@ -1,4 +1,4 @@
-﻿/*  =========================================================================
+/*  =========================================================================
     zframe - working with single message frames
 
     Copyright (c) the Contributors as noted in the AUTHORS file.
@@ -24,7 +24,7 @@
 @end
 */
 
-#include "../include/czmq.h"
+#include "czmq_classes.h"
 
 //  zframe_t instances always have this tag as the first 4 octets of
 //  their data, which lets us do runtime object typing & validation.
@@ -36,6 +36,10 @@ struct _zframe_t {
     uint32_t tag;               //  Object tag for runtime detection
     zmq_msg_t zmsg;             //  zmq_msg_t blob for frame
     int more;                   //  More flag, from last read
+    uint32_t routing_id;        //  Routing ID back to sender, if any
+#ifdef ZMQ_GROUP_MAX_LENGTH
+    char group[ZMQ_GROUP_MAX_LENGTH + 1];
+#endif
 };
 
 //  --------------------------------------------------------------------------
@@ -46,16 +50,20 @@ zframe_t *
 zframe_new (const void *data, size_t size)
 {
     zframe_t *self = (zframe_t *) zmalloc (sizeof (zframe_t));
-    if (self) {
-        self->tag = ZFRAME_TAG;
-        if (size) {
-            zmq_msg_init_size (&self->zmsg, size);
-            if (data)
-                memcpy (zmq_msg_data (&self->zmsg), data, size);
+    assert (self);
+    self->tag = ZFRAME_TAG;
+    if (size) {
+        //  Catch heap exhaustion in this specific case
+        if (zmq_msg_init_size (&self->zmsg, size)) {
+            zframe_destroy (&self);
+            return NULL;
         }
-        else
-            zmq_msg_init (&self->zmsg);
+        if (data)
+            memcpy (zmq_msg_data (&self->zmsg), data, size);
     }
+    else
+        zmq_msg_init (&self->zmsg);
+
     return self;
 }
 
@@ -68,10 +76,9 @@ zframe_t *
 zframe_new_empty (void)
 {
     zframe_t *self = (zframe_t *) zmalloc (sizeof (zframe_t));
-    if (self) {
-        self->tag = ZFRAME_TAG;
-        zmq_msg_init (&self->zmsg);
-    }
+    assert (self);
+    self->tag = ZFRAME_TAG;
+    zmq_msg_init (&self->zmsg);
     return self;
 }
 
@@ -88,7 +95,7 @@ zframe_destroy (zframe_t **self_p)
         assert (zframe_is (self));
         zmq_msg_close (&self->zmsg);
         self->tag = 0xDeadBeef;
-        free (self);
+        freen (self);
         *self_p = NULL;
     }
 }
@@ -116,13 +123,23 @@ zframe_recv (void *source)
     assert (source);
     void *handle = zsock_resolve (source);
     zframe_t *self = zframe_new (NULL, 0);
-    if (self) {
-        if (zmq_recvmsg (handle, &self->zmsg, 0) < 0) {
-            zframe_destroy (&self);
-            return NULL;            //  Interrupted or terminated
-        }
-        self->more = zsock_rcvmore (handle);
+    assert (self);
+
+    if (zmq_recvmsg (handle, &self->zmsg, 0) < 0) {
+        zframe_destroy (&self);
+        return NULL;            //  Interrupted or terminated
     }
+    self->more = zsock_rcvmore (source);
+#if defined (ZMQ_SERVER)
+    //  Grab routing ID if we're reading from a SERVER socket (ZMQ 4.2 and later)
+    if (zsock_type (source) == ZMQ_SERVER)
+        self->routing_id = zmq_msg_routing_id (&self->zmsg);
+#endif
+#if defined (ZMQ_DISH)
+    //  Grab group if we're reading from a DISH Socket (ZMQ 4.2 and later)
+    if (zsock_type (source) == ZMQ_DISH)
+        strcpy (self->group, zmq_msg_group (&self->zmsg));
+#endif
     return self;
 }
 
@@ -149,12 +166,28 @@ zframe_send (zframe_t **self_p, void *dest, int flags)
             zmq_msg_init (&copy);
             if (zmq_msg_copy (&copy, &self->zmsg))
                 return -1;
+#if defined (ZMQ_SERVER)
+            if (zsock_type (dest) == ZMQ_SERVER)
+                zmq_msg_set_routing_id (&copy, self->routing_id);
+#endif
+#if defined (ZMQ_RADIO)
+            if (zsock_type (dest) == ZMQ_RADIO)
+                zmq_msg_set_group (&copy, self->group);
+#endif
             if (zmq_sendmsg (handle, &copy, send_flags) == -1) {
                 zmq_msg_close (&copy);
                 return -1;
             }
         }
         else {
+#if defined (ZMQ_SERVER)
+            if (zsock_type (dest) == ZMQ_SERVER)
+                zmq_msg_set_routing_id (&self->zmsg, self->routing_id);
+#endif
+#if defined (ZMQ_RADIO)
+            if (zsock_type (dest) == ZMQ_RADIO)
+                zmq_msg_set_group (&self->zmsg, self->group);
+#endif
             if (zmq_sendmsg (handle, &self->zmsg, send_flags) >= 0)
                 zframe_destroy (self_p);
             else
@@ -164,6 +197,7 @@ zframe_send (zframe_t **self_p, void *dest, int flags)
     return 0;
 }
 
+
 //  --------------------------------------------------------------------------
 //  Return size of frame.
 
@@ -190,6 +224,25 @@ zframe_data (zframe_t *self)
 }
 
 
+//  --------------------------------------------------------------------------
+//  Return meta data property for frame.
+//  The caller shall not modify or free the returned value, which shall be
+//  owned by the message.
+
+const char *
+zframe_meta (zframe_t *self, const char *property)
+{
+#if (ZMQ_VERSION >= ZMQ_MAKE_VERSION (4, 1, 0))
+    assert (self);
+    assert (zframe_is (self));
+
+    return zmq_msg_gets (&self->zmsg, property);
+#else
+    return NULL;
+#endif
+}
+
+
 //  --------------------------------------------------------------------------
 //  Create a new frame that duplicates an existing frame. If frame is null,
 //  or memory was exhausted, returns null.
@@ -221,7 +274,7 @@ zframe_strhex (zframe_t *self)
 
     size_t size = zframe_size (self);
     byte *data = zframe_data (self);
-    char *hex_str = (char *) zmalloc (size * 2 + 1);
+    char *hex_str = (char *) malloc (size * 2 + 1);
     if (!hex_str)
         return NULL;
 
@@ -247,10 +300,9 @@ zframe_strdup (zframe_t *self)
 
     size_t size = zframe_size (self);
     char *string = (char *) malloc (size + 1);
-    if (string) {
-        memcpy (string, zframe_data (self), size);
-        string [size] = 0;
-    }
+    assert (string);
+    memcpy (string, zframe_data (self), size);
+    string [size] = 0;
     return string;
 }
 
@@ -301,6 +353,68 @@ zframe_set_more (zframe_t *self, int more)
 }
 
 
+//  --------------------------------------------------------------------------
+//  Return frame routing ID, if the frame came from a ZMQ_SERVER socket.
+//  Else returns zero.
+
+uint32_t
+zframe_routing_id (zframe_t *self)
+{
+    assert (self);
+    assert (zframe_is (self));
+    return self->routing_id;
+}
+
+
+//  --------------------------------------------------------------------------
+//  Set routing ID on frame. This is used if/when the frame is sent to a
+//  ZMQ_SERVER socket.
+
+void
+zframe_set_routing_id (zframe_t *self, uint32_t routing_id)
+{
+    assert (self);
+    assert (zframe_is (self));
+    self->routing_id = routing_id;
+}
+
+
+//  --------------------------------------------------------------------------
+//  Return frame group of radio-dish pattern.
+const char *
+zframe_group (zframe_t *self)
+{
+    assert (self);
+#ifdef ZMQ_DISH
+    return self->group;
+#else
+    return NULL;
+#endif
+}
+
+
+//  --------------------------------------------------------------------------
+//  Set group on frame. This is used if/when the frame is sent to a
+//  ZMQ_RADIO socket.
+//  Return -1 on error, 0 on success.
+int
+zframe_set_group (zframe_t *self, const char *group)
+{
+#ifdef ZMQ_RADIO
+    if (strlen(group) > ZMQ_GROUP_MAX_LENGTH) {
+        errno = EINVAL;
+        return -1;
+    }
+
+    strcpy (self->group, group);
+    return 0;
+#else
+    errno = EINVAL;
+    return -1;
+#endif
+}
+
+
 //  --------------------------------------------------------------------------
 //  Return true if two frames have identical size and data
 
@@ -402,13 +516,22 @@ zframe_recv_nowait (void *source)
     void *handle = zsock_resolve (source);
 
     zframe_t *self = zframe_new (NULL, 0);
-    if (self) {
-        if (zmq_recvmsg (handle, &self->zmsg, ZMQ_DONTWAIT) < 0) {
-            zframe_destroy (&self);
-            return NULL;            //  Interrupted or terminated
-        }
-        self->more = zsock_rcvmore (handle);
+    assert (self);
+    if (zmq_recvmsg (handle, &self->zmsg, ZMQ_DONTWAIT) < 0) {
+        zframe_destroy (&self);
+        return NULL;            //  Interrupted or terminated
     }
+    self->more = zsock_rcvmore (source);
+#if defined (ZMQ_SERVER)
+    //  Grab routing ID if we're reading from a SERVER socket (ZMQ 4.2 and later)
+    if (zsock_type (source) == ZMQ_SERVER)
+        self->routing_id = zmq_msg_routing_id (&self->zmsg);
+#endif
+#if defined (ZMQ_DISH)
+    //  Grab group if we're reading from a DISH Socket (ZMQ 4.2 and later)
+    if (zsock_type (source) == ZMQ_DISH)
+        strcpy (self->group, zmq_msg_group (&self->zmsg));
+#endif
     return self;
 }
 
@@ -464,10 +587,14 @@ zframe_test (bool verbose)
 
     //  @selftest
     //  Create two PAIR sockets and connect over inproc
-    zsock_t *output = zsock_new_pair ("@inproc://zframe.test");
+    zsock_t *output = zsock_new (ZMQ_PAIR);
     assert (output);
-    zsock_t *input = zsock_new_pair (">inproc://zframe.test");
+    int port = zsock_bind (output, "tcp://127.0.0.1:*");
+    assert (port != -1);
+    zsock_t *input = zsock_new (ZMQ_PAIR);
     assert (input);
+    rc = zsock_connect (input, "tcp://127.0.0.1:%d", port);
+    assert (rc != -1);
 
     //  Send five different frames, test ZFRAME_MORE
     int frame_nbr;
@@ -505,10 +632,10 @@ zframe_test (bool verbose)
     zframe_reset (frame, "END", 3);
     char *string = zframe_strhex (frame);
     assert (streq (string, "454E44"));
-    free (string);
+    freen (string);
     string = zframe_strdup (frame);
     assert (streq (string, "END"));
-    free (string);
+    freen (string);
     rc = zframe_send (&frame, output, 0);
     assert (rc == 0);
 
@@ -527,9 +654,103 @@ zframe_test (bool verbose)
     }
     assert (frame_nbr == 10);
 
+#if (ZMQ_VERSION >= ZMQ_MAKE_VERSION (4, 1, 0))
+    // Test zframe_meta
+    frame = zframe_new ("Hello", 5);
+    assert (frame);
+    rc = zframe_send (&frame, output, 0);
+    assert (rc == 0);
+    frame = zframe_recv (input);
+    const char *meta = zframe_meta (frame, "Socket-Type");
+    assert (meta != NULL);
+    assert (streq (meta, "PAIR"));
+    assert (zframe_meta (frame, "nonexistent") == NULL);
+    zframe_destroy (&frame);
+#endif
+
     zsock_destroy (&input);
     zsock_destroy (&output);
 
+#if defined (ZMQ_SERVER)
+    //  Create server and client sockets and connect over inproc
+    zsock_t *server = zsock_new_server ("inproc://zframe-test-routing");
+    assert (server);
+    zsock_t *client = zsock_new_client ("inproc://zframe-test-routing");
+    assert (client);
+
+    //  Send request from client to server
+    zframe_t *request = zframe_new ("Hello", 5);
+    assert (request);
+    rc = zframe_send (&request, client, 0);
+    assert (rc == 0);
+    assert (!request);
+
+    //  Read request and send reply
+    request = zframe_recv (server);
+    assert (request);
+    assert (zframe_streq (request, "Hello"));
+    assert (zframe_routing_id (request));
+
+    zframe_t *reply = zframe_new ("World", 5);
+    assert (reply);
+    zframe_set_routing_id (reply, zframe_routing_id (request));
+    rc = zframe_send (&reply, server, 0);
+    assert (rc == 0);
+    zframe_destroy (&request);
+
+    //  Read reply
+    reply = zframe_recv (client);
+    assert (zframe_streq (reply, "World"));
+    assert (zframe_routing_id (reply) == 0);
+    zframe_destroy (&reply);
+
+    //  Client and server disallow multipart
+    frame = zframe_new ("Hello", 5);
+    rc = zframe_send (&frame, client, ZFRAME_MORE);
+    assert (rc == -1);
+    rc = zframe_send (&frame, server, ZFRAME_MORE);
+    assert (rc == -1);
+    zframe_destroy (&frame);
+
+    zsock_destroy (&client);
+    zsock_destroy (&server);
+#endif
+
+#ifdef ZMQ_RADIO
+    //  Create radio and dish sockets and connect over inproc
+    zsock_t *radio = zsock_new_radio ("inproc://zframe-test-radio");
+    assert (radio);
+    zsock_t *dish = zsock_new_dish ("inproc://zframe-test-radio");
+    assert (dish);
+
+    //  Join the group
+    rc = zsock_join (dish, "World");
+    assert (rc == 0);
+
+    //  Publish message from radio
+    zframe_t *message = zframe_new ("Hello", 5);
+    assert (message);
+    rc = zframe_set_group (message, "World");
+    assert (rc == 0);
+    rc = zframe_send (&message, radio, 0);
+    assert (rc == 0);
+    assert (!message);
+
+    //  Receive the message from dish
+    message = zframe_recv (dish);
+    assert (message);
+    assert (zframe_streq (message, "Hello"));
+    assert (strcmp("World", zframe_group (message)) == 0);
+    zframe_destroy (&message);
+
+    zsock_destroy (&dish);
+    zsock_destroy (&radio);
+#endif
+
+#if defined (__WINDOWS__)
+    zsys_shutdown();
+#endif
+
     //  @end
     printf ("OK\n");
 }
