@@ -29,6 +29,7 @@
     * CONNECT endpoint -- connect the gossip service to the specified peer
     * PUBLISH key value -- publish a key/value pair to the gossip cluster
     * STATUS -- return number of key/value pairs held by gossip service
+    * ZAP DOMAIN domain -- set the ZAP DOMAIN domain = value
 
     Returns these messages:
 
@@ -76,9 +77,8 @@
     The basic logic of the gossip service is to accept PUBLISH messages
     from its owning application, and to forward these to every remote, and
     every client it talks to. When a node gets a duplicate tuple, it throws
-    it away. When a node gets a new tuple, it stores it, and fowards it as
-    just described. At any point the application can access the node's set
-    of tuples.
+    it away. When a node gets a new tuple, it stores it, and forwards it as
+    just described.
 
     At present there is no way to expire tuples from the network.
 
@@ -91,6 +91,11 @@
 
 #include "czmq_classes.h"
 
+// TODO- project.xml?
+#ifdef CZMQ_BUILD_DRAFT_API
+#define CZMQ_ZGOSSIP_ZAP_DOMAIN "global"
+#endif
+
 
 //  ---------------------------------------------------------------------
 //  Forward declarations for the two main classes we use here
@@ -115,6 +120,12 @@ struct _server_t {
 
     tuple_t *cur_tuple;         //  Holds current tuple to publish
     zgossip_msg_t *message;     //  Message to broadcast
+
+    char *public_key;
+    char *secret_key;
+#ifdef CZMQ_BUILD_DRAFT_API
+    char *zap_domain;
+#endif
 };
 
 //  ---------------------------------------------------------------------
@@ -144,9 +155,9 @@ static void
 tuple_free (void *argument)
 {
     tuple_t *self = (tuple_t *) argument;
-    free (self->key);
-    free (self->value);
-    free (self);
+    freen (self->key);
+    freen (self->value);
+    freen (self);
 }
 
 //  Handle traffic from remotes
@@ -168,13 +179,17 @@ server_initialize (server_t *self)
     //  override this with a SET message.
     engine_configure (self, "server/timeout", "1000");
     self->message = zgossip_msg_new ();
-    
+
     self->remotes = zlistx_new ();
     assert (self->remotes);
     zlistx_set_destructor (self->remotes, (czmq_destructor *) zsock_destroy);
-    
+
     self->tuples = zhashx_new ();
     assert (self->tuples);
+
+#ifdef CZMQ_BUILD_DRAFT_API
+    self->zap_domain = strdup(CZMQ_ZGOSSIP_ZAP_DOMAIN);
+#endif
     return 0;
 }
 
@@ -186,20 +201,43 @@ server_terminate (server_t *self)
     zgossip_msg_destroy (&self->message);
     zlistx_destroy (&self->remotes);
     zhashx_destroy (&self->tuples);
+    zstr_free (&self->public_key);
+    zstr_free (&self->secret_key);
+#ifdef CZMQ_BUILD_DRAFT_API
+    zstr_free (&self->zap_domain);
+#endif
 }
 
 //  Connect to a remote server
-
 static void
+#ifdef CZMQ_BUILD_DRAFT_API
+server_connect (server_t *self, const char *endpoint, const char *public_key)
+#else
 server_connect (server_t *self, const char *endpoint)
+#endif
 {
     zsock_t *remote = zsock_new (ZMQ_DEALER);
     assert (remote);          //  No recovery if exhausted
 
+#ifdef CZMQ_BUILD_DRAFT_API
+    if (public_key){
+        zcert_t *cert = zcert_new_from_txt (self->public_key, self->secret_key);
+        zcert_apply(cert, remote);
+        zsock_set_curve_serverkey (remote, public_key);
+#ifndef ZMQ_CURVE
+        // legacy ZMQ support
+        // inline incase the underlying assert is removed
+        bool ZMQ_CURVE = false;
+#endif
+        assert (zsock_mechanism (remote) == ZMQ_CURVE);
+        zcert_destroy(&cert);
+    }
+#endif
     //  Never block on sending; we use an infinite HWM and buffer as many
     //  messages as needed in outgoing pipes. Note that the maximum number
     //  is the overall tuple set size.
     zsock_set_unbounded (remote);
+
     if (zsock_connect (remote, "%s", endpoint)) {
         zsys_warning ("bad zgossip endpoint '%s'", endpoint);
         zsock_destroy (&remote);
@@ -209,7 +247,7 @@ server_connect (server_t *self, const char *endpoint)
     zgossip_msg_t *gossip = zgossip_msg_new ();
     zgossip_msg_set_id (gossip, ZGOSSIP_MSG_HELLO);
     zgossip_msg_send (gossip, remote);
-    
+
     tuple_t *tuple = (tuple_t *) zhashx_first (self->tuples);
     while (tuple) {
         zgossip_msg_set_id (gossip, ZGOSSIP_MSG_PUBLISH);
@@ -275,7 +313,16 @@ server_method (server_t *self, const char *method, zmsg_t *msg)
     if (streq (method, "CONNECT")) {
         char *endpoint = zmsg_popstr (msg);
         assert (endpoint);
+#ifdef CZMQ_BUILD_DRAFT_API
+        // leaving this in here for now because if/def changes the server_connect
+        // function args. it doesn't look like server_connect is used anywhere else
+        // but want to leave this in until we're sure this is stable..
+        char *public_key = zmsg_popstr (msg);
+        server_connect (self, endpoint, public_key);
+        zstr_free (&public_key);
+#else
         server_connect (self, endpoint);
+#endif
         zstr_free (&endpoint);
     }
     else
@@ -294,12 +341,45 @@ server_method (server_t *self, const char *method, zmsg_t *msg)
         zmsg_addstr (reply, "STATUS");
         zmsg_addstrf (reply, "%d", (int) zhashx_size (self->tuples));
     }
+#ifdef CZMQ_BUILD_DRAFT_API
+    else
+    if (streq (method, "SET PUBLICKEY")) {
+        char *key = zmsg_popstr (msg);
+        self->public_key = strdup (key);
+        assert (self->public_key);
+        zstr_free (&key);
+    }
+    else
+    if (streq (method, "SET SECRETKEY")) {
+        char *key = zmsg_popstr (msg);
+        self->secret_key = strdup(key);
+        assert (self->secret_key);
+        zstr_free (&key);
+    }
+    else
+    if (streq (method, "ZAP DOMAIN")) {
+        char *value = zmsg_popstr (msg);
+        zstr_free(&self->zap_domain);
+        self->zap_domain = strdup(value);
+        assert (self->zap_domain);
+        zstr_free (&value);
+    }
+#endif
     else
         zsys_error ("unknown zgossip method '%s'", method);
 
     return reply;
 }
 
+//  Apply new configuration.
+
+static void
+server_configuration (server_t *self, zconfig_t *config)
+{
+    ZPROTO_UNUSED(self);
+    ZPROTO_UNUSED(config);
+    //  Apply new configuration
+}
 
 //  Allocate properties and structures for a new client connection and
 //  optionally engine_set_next_event (). Return 0 if OK, or -1 on error.
@@ -436,7 +516,7 @@ zgossip_test (bool verbose)
     zsock_set_rcvtimeo (client, 2000);
     int rc = zsock_connect (client, "inproc://zgossip");
     assert (rc == 0);
-    
+
     //  Send HELLO, which gets no message
     zgossip_msg_t *message = zgossip_msg_new ();
     zgossip_msg_set_id (message, ZGOSSIP_MSG_HELLO);
@@ -448,7 +528,7 @@ zgossip_test (bool verbose)
     zgossip_msg_recv (message, client);
     assert (zgossip_msg_id (message) == ZGOSSIP_MSG_PONG);
     zgossip_msg_destroy (&message);
-    
+
     zactor_destroy (&server);
     zsock_destroy (&client);
 
@@ -476,13 +556,134 @@ zgossip_test (bool verbose)
     //  got nothing
     zclock_sleep (200);
 
+    zstr_send (alpha, "STATUS");
+    char *command, *status, *key, *value;
+
+    zstr_recvx (alpha, &command, &key, &value, NULL);
+    assert (streq (command, "DELIVER"));
+    assert (streq (key, "inproc://alpha-1"));
+    assert (streq (value, "service1"));
+    zstr_free (&command);
+    zstr_free (&key);
+    zstr_free (&value);
+
+    zstr_recvx (alpha, &command, &key, &value, NULL);
+    assert (streq (command, "DELIVER"));
+    assert (streq (key, "inproc://alpha-2"));
+    assert (streq (value, "service2"));
+    zstr_free (&command);
+    zstr_free (&key);
+    zstr_free (&value);
+
+    zstr_recvx (alpha, &command, &key, &value, NULL);
+    assert (streq (command, "DELIVER"));
+    assert (streq (key, "inproc://beta-1"));
+    assert (streq (value, "service1"));
+    zstr_free (&command);
+    zstr_free (&key);
+    zstr_free (&value);
+
+    zstr_recvx (alpha, &command, &key, &value, NULL);
+    assert (streq (command, "DELIVER"));
+    assert (streq (key, "inproc://beta-2"));
+    assert (streq (value, "service2"));
+    zstr_free (&command);
+    zstr_free (&key);
+    zstr_free (&value);
+
+    zstr_recvx (alpha, &command, &status, NULL);
+    assert (streq (command, "STATUS"));
+    assert (atoi (status) == 4);
+    zstr_free (&command);
+    zstr_free (&status);
+
     zactor_destroy (&base);
     zactor_destroy (&alpha);
     zactor_destroy (&beta);
 
+#ifdef CZMQ_BUILD_DRAFT_API
+    // curve
+    if (zsys_has_curve()) {
+        if (verbose)
+            printf("testing CURVE support");
+        zclock_sleep (2000);
+        zactor_t *auth = zactor_new(zauth, NULL);
+        assert (auth);
+        if (verbose) {
+            zstr_sendx (auth, "VERBOSE", NULL);
+            zsock_wait (auth);
+        }
+        zstr_sendx(auth,"ALLOW","127.0.0.1",NULL);
+        zsock_wait(auth);
+        zstr_sendx (auth, "CURVE", CURVE_ALLOW_ANY, NULL);
+        zsock_wait (auth);
+
+        server = zactor_new (zgossip, "server");
+        if (verbose)
+            zstr_send (server, "VERBOSE");
+        assert (server);
+
+        zcert_t *client1_cert = zcert_new ();
+        zcert_t *server_cert = zcert_new ();
+
+        zstr_sendx (server, "SET PUBLICKEY", zcert_public_txt (server_cert), NULL);
+        zstr_sendx (server, "SET SECRETKEY", zcert_secret_txt (server_cert), NULL);
+        zstr_sendx (server, "ZAP DOMAIN", "TEST", NULL);
+
+        zstr_sendx (server, "BIND", "tcp://127.0.0.1:*", NULL);
+        zstr_sendx (server, "PORT", NULL);
+        zstr_recvx (server, &command, &value, NULL);
+        assert (streq (command, "PORT"));
+        int port = atoi (value);
+        zstr_free (&command);
+        zstr_free (&value);
+        char endpoint [32];
+        sprintf (endpoint, "tcp://127.0.0.1:%d", port);
+
+        zactor_t *client1 = zactor_new (zgossip, "client");
+        if (verbose)
+            zstr_send (client1, "VERBOSE");
+        assert (client1);
+
+        zstr_sendx (client1, "SET PUBLICKEY", zcert_public_txt (client1_cert), NULL);
+        zstr_sendx (client1, "SET SECRETKEY", zcert_secret_txt (client1_cert), NULL);
+        zstr_sendx (client1, "ZAP DOMAIN", "TEST", NULL);
+
+        const char *public_txt = zcert_public_txt (server_cert);
+        zstr_sendx (client1, "CONNECT", endpoint, public_txt, NULL);
+        zstr_sendx (client1, "PUBLISH", "tcp://127.0.0.1:9001", "service1", NULL);
+
+        zclock_sleep (500);
+
+        zstr_send (server, "STATUS");
+        zclock_sleep (500);
+
+        zstr_recvx (server, &command, &key, &value, NULL);
+        assert (streq (command, "DELIVER"));
+        assert (streq (value, "service1"));
+
+        zstr_free (&command);
+        zstr_free (&key);
+        zstr_free (&value);
+
+        zstr_sendx (client1, "$TERM", NULL);
+        zstr_sendx (server, "$TERM", NULL);
+
+        zclock_sleep(500);
+
+        zcert_destroy (&client1_cert);
+        zcert_destroy (&server_cert);
+
+        zactor_destroy (&client1);
+        zactor_destroy (&server);
+        zactor_destroy (&auth);
+    }
+#endif
+
+#if defined (__WINDOWS__)
+    zsys_shutdown();
+#endif
+
     //  @end
     printf ("OK\n");
-
-    // zgossip_msg is a private class used by zgossip, so we run the test here
-    zgossip_msg_test (verbose);
 }
